"""Utility functions for logging."""
import collections
import functools
import json
import logging.config
import os


def configure_logger():
    """
    Configures the logger.

    If a configuration file does not exist in the root user directory, then
    default configuration options will be written to file.  This file can be
    customized and will be read on the next configuration.

    """
    # Attempt to read SDK logging config file
    config_file = os.path.join(os.path.expanduser('~'),
                               'helios_logger_config.json')
    config = None
    if os.path.exists(config_file):
        try:
            with open(config_file, 'r') as f:
                config = json.load(f)
        except ValueError:
            config = None

    if config is None:
        # Default logging configuration.
        config = collections.defaultdict(dict)

        config['version'] = 1
        config['disable_existing_loggers'] = 1

        config['formatters']['simple'] = {'format': '%(asctime)s-%(levelname)s'
                                                    '-%(module)s-%(name)s'
                                                    '-%(funcName)s: %(message)s',
                                          'datefmt': '%H:%M:%S'}

        config['handlers']['console'] = {'class': 'logging.StreamHandler',
                                         'level': 'WARNING',
                                         'formatter': 'simple',
                                         'stream': 'ext://sys.stdout'}

        config['root'] = {'level': 'INFO',
                          'handlers': ['console']}

        # Write default configuration to file
        try:
            with open(config_file, 'w') as f:
                json.dump(config, f)
        except IOError:
            pass

    logging.config.dictConfig(config)


def log_entrance_exit(func):
    """
    Decorator to log the entrance and exit of method calls.

    This will make use of the logger instance passed via self.  When using
    functools.wraps, self will be the first index in args.
    See self = args[0] comment below.

    Args:
        func: The function to be wrapped.

    Returns:
        Wrapped function.

    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # self should be first argument.
        func_self = args[0]
        if len(args) > 1:
            func_self._logger.info('Entering %s: %s, %s', func.__name__, args[1:], kwargs)
        else:
            func_self._logger.info('Entering %s: %s', func.__name__, kwargs)

        # Evaluate wrapped function.
        f_result = func(*args, **kwargs)

        if len(args) > 1:
            func_self._logger.info('Exiting %s: %s, %s', func.__name__, args[1:], kwargs)
        else:
            func_self._logger.info('Exiting %s: %s', func.__name__, kwargs)
        return f_result

    return wrapper



"""Mixins and core functionality."""
import os
from collections import namedtuple
from contextlib import closing
from io import BytesIO
from math import ceil
from multiprocessing.pool import ThreadPool

import numpy as np
import requests
from PIL import Image

from helios.core.request_manager import RequestManager
from helios.core.session import Session
from helios.core.structure import ImageRecord, Record
from helios.utilities import logging_utils, parsing_utils


class SDKCore(object):
    """
    Core class for Python interface to Helios Core APIs.

    This class must be inherited by any additional Core API classes.
    """
    _max_threads = 32

    def __init__(self, session=None):
        """
        Initialize core API instance.

        If a session has been started prior to initialization it can be used
        via the session input parameter. If this is not used a session will
        be started automatically.

        Args:
            session (helios.Session object, optional): An instance of the
                Session. Defaults to None. If unused a session will be
                created for you.

        """
        # Start session or use custom session.
        if session is None:
            self._session = Session()
        else:
            self._session = session

        # If the session hasn't been started, start it.
        if not self._session.token:
            self._session.start_session()

        # Create request manager to handle all API requests.
        self._request_manager = RequestManager(self._session.token,
                                               pool_maxsize=self._max_threads)

    @property
    def _base_api_url(self):
        return self._session.api_url

    @_base_api_url.setter
    def _base_api_url(self, value):
        raise AttributeError('Access to _base_api_url is restricted.')

    @staticmethod
    def _parse_query_inputs(parameters):
        """
        Create query string from a dictionary of parameters.

        Args:
            parameters (dict):  Key/values to combine into a query string.

        Returns:
            str: Query string.
        """
        parameters_temp = parameters.copy()

        query_str = ''

        # Check for unique case: sensors
        if 'sensors' in parameters_temp:
            query_str += parameters_temp.pop('sensors') + '&'

        # Parse input_dict into a str
        for key, val in parameters_temp.items():
            if val is None:
                continue

            if isinstance(val, (list, tuple)):
                query_str += (str(key) +
                              '=' +
                              ','.join([str(x) for x in val]) +
                              '&')
            else:
                query_str += (str(key) + '=' + str(val) + '&')

        # Remove final ampersand
        query_str = query_str[:-1]

        return query_str

    def _process_messages(self, func, messages):
        # Create thread pool
        with closing(ThreadPool(self._max_threads)) as thread_pool:
            results = thread_pool.map(func, messages)

        # Try to determine how many were successful.
        try:
            n_messages = len(messages)
            n_successful = sum([True for x in results if x.ok])
            log_message = '{} out of {} successful'.format(n_successful, n_messages)
            if n_successful == 0:
                self._logger.error(log_message)
            elif n_successful < n_messages:
                self._logger.warning(log_message)
            else:
                self._logger.info(log_message)
        except AttributeError:
            pass

        return results


class IndexMixin(object):
    """Mixin for index queries."""

    @logging_utils.log_entrance_exit
    def index(self, **kwargs):
        max_skip = 4000

        limit = int(kwargs.pop('limit', 100))
        skip = int(kwargs.pop('skip', 0))

        # Raise right away if skip is too high.
        if skip > max_skip:
            raise ValueError('skip must be less than the maximum skip value '
                             'of {}. A value of {} was tried.'.format(max_skip, skip))

        # Create the messages up to the maximum skip.
        Message = namedtuple('Message', ['kwargs', 'limit', 'skip'])

        messages = []
        for i in range(skip, max_skip, limit):
            if i + limit > max_skip:
                temp_limit = max_skip - i
            else:
                temp_limit = limit
            messages.append(Message(kwargs=kwargs, limit=temp_limit, skip=i))

        # Process first message to get the total number of features to expect.
        initial_resp = self.__index_worker(messages.pop(0))

        try:
            total = initial_resp.content['properties']['total']
        except KeyError:
            total = initial_resp.content['total']

        # If only one query was necessary, return immediately.
        if total < limit:
            return [initial_resp]

        # Warn the user when truncation occurs. (max_skip is hit)
        if total > max_skip:
            self._logger.warning('Maximum skip level. Truncated results for: %s',
                                 kwargs)

        # Determine number of iterations that will be needed.
        n_queries_needed = int(ceil((total - skip) / float(limit))) - 1
        messages = messages[0:n_queries_needed]

        # Log number of queries required.
        self._logger.info('%s index queries required for: %s', n_queries_needed, kwargs)

        # Process messages using the worker function.
        results = self._process_messages(self.__index_worker, messages)

        # Put initial query back in list.
        results.insert(0, initial_resp)

        return results

    def __index_worker(self, msg):
        """msg must contain kwargs (search criteria), limit, and skip"""

        params_str = self._parse_query_inputs(msg.kwargs)

        query_str = '{}/{}?{}&limit={}&skip={}'.format(self._base_api_url,
                                                       self._core_api,
                                                       params_str,
                                                       msg.limit,
                                                       msg.skip)

        # Perform query
        try:
            resp = self._request_manager.get(query_str)
        except requests.exceptions.RequestException as e:
            return Record(message=msg, query=query_str, error=e)

        return Record(message=msg, query=query_str, content=resp.json())


class ShowMixin(object):
    """Mixin for show queries"""

    @logging_utils.log_entrance_exit
    def show(self, ids):
        if not isinstance(ids, (list, tuple)):
            ids = [ids]

        # Create messages for worker.
        Message = namedtuple('Message', 'id_')
        messages = [Message(x) for x in ids]

        # Process messages using the worker function.
        results = self._process_messages(self.__show_worker, messages)

        return results

    def __show_worker(self, msg):
        """msg must contain id_"""
        query_str = '{}/{}/{}'.format(self._base_api_url, self._core_api, msg.id_)

        try:
            resp = self._request_manager.get(query_str)
        except requests.exceptions.RequestException as e:
            return Record(message=msg, query=query_str, error=e)

        return Record(message=msg, query=query_str, content=resp.json())


class ShowImageMixin(object):
    """Mixin for show_image queries"""

    @logging_utils.log_entrance_exit
    def show_image(self, id_, data, out_dir=None, return_image_data=False):
        if not isinstance(data, (list, tuple)):
            data = [data]

        # Create messages for worker.
        Message = namedtuple('Message', ['id_', 'data', 'out_dir', 'return_image_data'])
        messages = [Message(id_, x, out_dir, return_image_data) for x in data]

        # Make sure directory exists.
        if out_dir:
            if not os.path.exists(out_dir):
                os.makedirs(out_dir)

        # Process messages using the worker function.
        results = self._process_messages(self.__show_image_worker, messages)

        return results

    def __show_image_worker(self, msg):
        """msg must contain id_, data, out_dir, and return_image_data"""
        query_str = '{}/{}/{}/images/{}'.format(self._base_api_url,
                                                self._core_api,
                                                msg.id_,
                                                msg.data)

        try:
            resp = self._request_manager.get(query_str)
        except requests.exceptions.RequestException as e:
            return ImageRecord(message=msg, query=query_str, error=e)

        # Parse key from url.
        parsed_url = parsing_utils.parse_url(resp.url)
        _, image_name = os.path.split(parsed_url.path)

        # Read image from response.
        img = Image.open(BytesIO(resp.content))

        # Write image to file.
        if msg.out_dir is not None:
            out_file = os.path.join(msg.out_dir, image_name)
            img.save(out_file)
        else:
            out_file = None

        # Read and return image data.
        if msg.return_image_data:
            img_data = np.array(img)
        else:
            img_data = None

        return ImageRecord(message=msg, query=query_str, name=image_name,
                           content=img_data, output_file=out_file)



"""
Helios Cameras API.

Methods are meant to represent the core functionality in the developer
documentation.  Some may have additional functionality for convenience.

"""
import logging

from dateutil.parser import parse
from helios.core.mixins import SDKCore, ShowMixin, ShowImageMixin, IndexMixin
from helios.core.structure import RecordCollection
from helios.utilities import logging_utils


class Cameras(ShowImageMixin, ShowMixin, IndexMixin, SDKCore):
    """The Cameras API provides access to all cameras in the Helios Network."""

    _core_api = 'cameras'

    def __init__(self, session=None):
        """
        Initialize Cameras instance.

        Args:
            session (helios.Session object, optional): An instance of the
                Session. Defaults to None. If unused a session will be
                created for you.

        """
        super(Cameras, self).__init__(session=session)
        self._logger = logging.getLogger(__name__)

    @logging_utils.log_entrance_exit
    def images(self, camera_id, start_time, end_time=None, limit=500):
        """
        Get the image times available for a given camera in the media cache.

        The media cache contains all recent images archived by Helios, either
        for internal analytics or for end user recording purposes.

        Args:
            camera_id (str): Camera ID.
            start_time (str): Starting image timestamp, specified in UTC as an
                ISO 8601 string (e.g. 2014-08-01 or 2014-08-01T12:34:56.000Z).
            end_time (str, optional): Ending image timestamp, specified in UTC
                as an ISO 8601 string (e.g. 2014-08-01 or 2014-08-01T12:34:56.000Z).
            limit (int, optional): Number of images to be returned, up to a max
                of 500. Defaults to 500.

        Returns:
            sequence of strs: Image times.

        """
        if end_time:
            end = parse(end_time).utctimetuple()
        else:
            end = None

        image_times = []
        while True:
            query_str = '{}/{}/{}/images?time={}&limit={}'.format(self._base_api_url,
                                                                  self._core_api,
                                                                  camera_id,
                                                                  start_time,
                                                                  limit)
            # Get image times available.
            resp = self._request_manager.get(query_str)
            times = resp.json()['times']

            # Return now if no end_time was provided.
            if end_time is None:
                image_times.extend(times)
                break

            # Parse the last time and break if no times were found
            try:
                last = parse(times[-1]).utctimetuple()
            except IndexError:
                break

            # the last image is still newer than the end time, keep looking
            if last < end:
                if len(times) > 1:
                    image_times.extend(times[0:-1])
                    start_time = times[-1]
                else:
                    image_times.extend(times)
                    break
            # The end time is somewhere in between.
            elif last > end:
                good_times = [x for x in times if parse(x).utctimetuple() < end]
                image_times.extend(good_times)
                break
            else:
                image_times.extend(times)
                break

        if not image_times:
            self._logger.warning('No images were found for %s in the %s to %s range.',
                                 camera_id, start_time, end_time)

        return image_times

    def index(self, **kwargs):
        """
        Get cameras matching the provided spatial, text, or
        metadata filters.

        The maximum skip value is 4000. If this is reached, truncated results
        will be returned. You will need to refine your query to avoid this.

        .. _cameras_index_documentation: https://helios.earth/developers/api/cameras/#index

        Args:
            **kwargs: Any keyword arguments found in the cameras_index_documentation_.

        Returns:
             :class:`IndexResults <helios.cameras_api.IndexResults>`

        """
        results = super(Cameras, self).index(**kwargs)

        content = []
        for record in results:
            if record.ok:
                for feature in record.content['features']:
                    content.append(CamerasFeature(feature))

        return IndexResults(content, results)

    def show(self, camera_ids):
        """
        Get attributes for cameras.

        Args:
            camera_ids (str or sequence of strs): Helios camera ID(s).

        Returns:
            :class:`ShowResults <helios.cameras_api.ShowResults>`

        """
        results = super(Cameras, self).show(camera_ids)

        content = []
        for record in results:
            if record.ok:
                content.append(CamerasFeature(record.content))

        return ShowResults(content, results)

    def show_image(self, camera_id, times, out_dir=None, return_image_data=False):
        """
        Get images from the media cache.

        The media cache contains all recent images archived by Helios, either
        for internal analytics or for end user recording purposes.

        Args:
            camera_id (str): Camera ID.
            times (str or sequence of strs): Image times, specified in UTC as
                an ISO 8601 string (e.g. 2017-08-01 or 2017-08-01T12:34:56.000Z).
                The image with the closest matching timestamp will be returned.
            out_dir (optional, str): Directory to write images to.  Defaults to
                None.
            return_image_data (optional, bool): If True images will be returned
                as numpy.ndarrays.  Defaults to False.

        Returns:
            :class:`ShowImageResults <helios.cameras_api.ShowImageResults>`

        """
        results = super(Cameras, self).show_image(camera_id,
                                                  times,
                                                  out_dir=out_dir,
                                                  return_image_data=return_image_data)

        content = []
        for record in results:
            if record.ok:
                content.append(record.content)

        return ShowImageResults(content, results)


class CamerasFeature(object):
    """
    Individual Camera GeoJSON feature.

    Attributes:
        city (str): 'city' value for the feature.
        country (str): 'country' value for the feature.
        description (str): 'description' value for the feature.
        direction (str): 'direction' value for the feature.
        id (str): 'id' value for the feature.
        json (dict): Raw 'json' for the feature.
        region (str): 'region' value for the feature.
        state (str): 'state' value for the feature.
        video (bool): 'video' value for the feature.

    """

    def __init__(self, feature):
        self.json = feature

        # Use dict.get built-in to guarantee all values will be initialized.
        self.city = feature['properties'].get('city')
        self.country = feature['properties'].get('country')
        self.description = feature['properties'].get('description')
        self.direction = feature['properties'].get('direction')
        self.id = feature.get('id')
        self.region = feature['properties'].get('region')
        self.state = feature['properties'].get('state')
        self.video = feature['properties'].get('video')


class CamerasFeatureCollection(RecordCollection):
    """Derived class for Cameras feature collections."""

    def __init__(self, content, records):
        super(CamerasFeatureCollection, self).__init__(content, records)

    @property
    def city(self):
        """'city' values for every feature."""
        return [x.city for x in self._content]

    @property
    def country(self):
        """'country' values for every feature."""
        return [x.country for x in self._content]

    @property
    def description(self):
        """'description' values for every feature."""
        return [x.description for x in self._content]

    @property
    def direction(self):
        """'direction' values for every feature."""
        return [x.direction for x in self._content]

    @property
    def id(self):
        """'id' values for every feature."""
        return [x.id for x in self._content]

    @property
    def json(self):
        """Raw 'json' for every feature."""
        return [x.json for x in self._content]

    @property
    def region(self):
        """'region' values for every feature."""
        return [x.region for x in self._content]

    @property
    def state(self):
        """'state' values for every feature."""
        return [x.state for x in self._content]

    @property
    def video(self):
        """'video' values for every feature."""
        return [x.video for x in self._content]


class IndexResults(CamerasFeatureCollection):
    """
    Index results for the Cameras API.

    IndexResults is an iterable for GeoJSON features.  This allows the
    user to iterate and select based on Feature attributes for each element.

    All features within IndexResults are instances of
    :class:`CamerasFeature <helios.cameras_api.CamerasFeature>`

    """

    def __init__(self, content, records):
        super(IndexResults, self).__init__(content, records)


class ShowImageResults(RecordCollection):
    """
    Show_image results for the Cameras API.

    ShowImageResults is an iterable for the fetched image content. Each element
    of the iterable will be an ndarray if return_image_data was True.

    """

    def __init__(self, content, records):
        super(ShowImageResults, self).__init__(content, records)

    @property
    def image_data(self):
        """Image data if return_image_data was True."""
        return self._content

    @property
    def output_file(self):
        """Full paths to all written images."""
        return [x.output_file for x in self._raw if x.ok]

    @property
    def name(self):
        """Names of all images."""
        return [x.name for x in self._raw if x.ok]


class ShowResults(CamerasFeatureCollection):
    """
    Show results for the Cameras API.

    ShowResults is an iterable for GeoJSON features.  This allows the
    user to iterate and select based on Feature attributes for each element.

    All features within ShowResults are instances of
    :class:`CamerasFeature <helios.cameras_api.CamerasFeature>`

    """

    def __init__(self, content, records):
        super(ShowResults, self).__init__(content, records)



"""
Helios Collections API.

Methods are meant to represent the core functionality in the developer
documentation.  Some may have additional functionality for convenience.

"""

import hashlib
import logging
from collections import namedtuple

import requests
from helios.core.mixins import SDKCore, IndexMixin, ShowImageMixin
from helios.core.structure import Record, RecordCollection
from helios.utilities import logging_utils


class Collections(ShowImageMixin, IndexMixin, SDKCore):
    """
    The Collections API allows users to group and organize individual image
    frames.

    Collections are intended to be short-lived resources and will be accessible
    for 90 days from the time the collection was created. After that time
    period has expired, the collection and all associated imagery will be
    removed from the system.

    """
    _core_api = 'collections'

    def __init__(self, session=None):
        """
        Initialize Collection instance.

        Args:
            session (helios.Session object, optional): An instance of the
                Session. Defaults to None. If unused a session will be
                created for you.

        """
        super(Collections, self).__init__(session=session)
        self._logger = logging.getLogger(__name__)

    @logging_utils.log_entrance_exit
    def add_image(self, collection_id, assets):
        """
        Add images to a collection from Helios assets.

        `assets` dictionary templates:

        .. code-block:: python

            # Asset examples that can be included in the `assets` input list.
            {'camera_id': ''}
            {'camera_id': '', 'time': ''}
            {'observation_id': ''}
            {'collection_is': '', 'image': ''}

        Usage example:

        .. code-block:: python

            import helios
            collections = helios.Collections()
            camera_id = '...'
            times = [...] # List of image times.
            destination_id = '...'
            data = [{'camera_id': camera_id, 'time': x} for x in times]
            collections.add_image(destination_id, data)

        Args:
            collection_id (str): Collection ID.
            assets (dict or sequence of dicts): Data containing any of these
                payloads (camera_id), (camera_id, time), (observation_id),
                (collection_id, image). E.g. data =
                [{'camera_id': 'cam_01', time: '2017-01-01T00:00:000Z'}]

        Returns:
            :class:`AddImageResults <helios.collections_api.AddImageResults>`

        """
        assert isinstance(assets, (list, tuple, dict))

        if isinstance(assets, dict):
            assets = [assets]

        # Create messages for worker.
        Message = namedtuple('Message', ['collection_id', 'data'])
        messages = [Message(collection_id, x) for x in assets]

        # Process messages using the worker function.
        results = self._process_messages(self.__add_image_worker, messages)

        return AddImageResults([x.content for x in results], results)

    def __add_image_worker(self, msg):
        """msg must contain collection_id and data"""
        # need to strip out the Bearer to work with a POST for collections
        post_token = self._request_manager.auth_token['value'].replace('Bearer ', '')

        # Compose post request
        parms = {'access_token': post_token}
        parms.update(msg.data)

        header = {'name': 'Content-Type',
                  'value': 'application/x-www-form-urlencoded'}
        post_url = '{}/collections/{}/images'.format(self._base_api_url,
                                                     msg.collection_id)

        try:
            resp = self._request_manager.post(post_url, headers=header, data=parms)
        except requests.exceptions.RequestException as e:
            return Record(message=msg, query=post_url, error=e)

        return Record(message=msg, query=post_url, content=resp.json())

    @logging_utils.log_entrance_exit
    def copy(self, collection_id, new_name):
        """
        Copy a collection and its contents to a new collection.

        Args:
            collection_id (str): Collection ID.
            new_name (str): New collection name.

        Returns:
            str: New collection ID.

        """
        # Get the collection metadata that needs to be copied.
        query_str = '{}/{}/{}'.format(self._base_api_url,
                                      self._core_api,
                                      collection_id)
        metadata = self._request_manager.get(query_str).json()

        # Get the images that exist in the collection.
        image_names = self.images(collection_id)

        # Create new collection.
        new_id = self.create(new_name, metadata['description'], metadata['tags'])

        # Add images to new collection.
        data = [{'collection_id': collection_id, 'image': x} for x in image_names]
        _ = self.add_image(new_id, data)

        return new_id

    @logging_utils.log_entrance_exit
    def create(self, name, description, tags=None):
        """
        Create a new collection.

        Args:
            name (str): Display name for the collection.
            description (str): Description for the collection.
            tags (str or sequence of strs, optional): Optional comma-delimited
                keyword tags to be added to the collection.

        Returns:
            str: New collection ID.

        """
        # need to strip out the Bearer to work with a POST for collections
        post_token = self._request_manager.auth_token['value'].replace('Bearer ', '')

        # Compose parms block
        parms = {'name': name, 'description': description, 'access_token': post_token}
        if tags is not None:
            if isinstance(tags, (list, tuple)):
                tags = ','.join(tags)
            parms['tags'] = tags

        header = {'name': 'Content-Type',
                  'value': 'application/x-www-form-urlencoded'}

        post_url = '{}/{}'.format(self._base_api_url, self._core_api)

        resp = self._request_manager.post(post_url, headers=header, data=parms).json()

        return resp['collection_id']

    @logging_utils.log_entrance_exit
    def delete(self, collection_id):
        """
        Delete an empty collection.

        If the collection is not empty, delete will fail.  Use the
        :meth:`empty <helios.collections_api.Collections.empty>` method to
        remove all imagery before calling this method.

        Args:
            collection_id (str): Collection to delete.

        Returns:
            dict: JSON response

        """
        query_str = '{}/{}/{}'.format(self._base_api_url,
                                      self._core_api,
                                      collection_id)

        resp = self._request_manager.delete(query_str)

        return resp.json()

    @logging_utils.log_entrance_exit
    def empty(self, collection_id):
        """
        Bulk remove (up to 1000) images from a collection.

        Args:
            collection_id (str): Collection to empty.

        Returns:
            dict: JSON response.

        """
        query_str = '{}/{}/{}/images'.format(self._base_api_url,
                                             self._core_api,
                                             collection_id)

        resp = self._request_manager.delete(query_str)

        return resp.json()

    @logging_utils.log_entrance_exit
    def images(self, collection_id, camera=None, old_flag=False):
        """
        Get all image names in a given collection.

        When using the optional camera input parameter only images from that
        camera will be returned.

        Args:
            collection_id (str): Collection ID.
            camera (str, optional): Camera ID to be found.
            old_flag (bool, optional): Flag for finding old format image names.
                When True images that do not contain md5 hashes at the start of
                their name will be found.

        Returns:
            sequence of strs: Image names.

        """
        mark_img = ''

        if camera is not None:
            if not old_flag:
                md5_str = hashlib.md5(camera.encode('utf-8')).hexdigest()
                camera = md5_str[0:4] + '-' + camera
            mark_img = camera

        good_images = []
        while True:
            results = self.show(collection_id, marker=mark_img)

            # Gather images.
            images_found = results.images

            if camera is not None:
                imgs_found_temp = [x for x in images_found if x.split('_')[0] == camera]
            else:
                imgs_found_temp = images_found

            if not imgs_found_temp:
                break

            good_images.extend(imgs_found_temp)
            if len(imgs_found_temp) < len(images_found):
                break
            else:
                mark_img = good_images[-1]

        return good_images

    def index(self, **kwargs):
        """
        Get collections matching the provided spatial, text, or metadata filters.

        The maximum skip value is 4000. If this is reached, truncated results
        will be returned. You will need to refine your query to avoid this.

        .. _collections_index_documentation: https://helios.earth/developers/api/collections/#index

        Args:
            **kwargs: Any keyword arguments found in the collections_index_documentation_.

        Returns:
             :class:`IndexResults <helios.collections_api.IndexResults>`

        """
        results = super(Collections, self).index(**kwargs)

        content = []
        for record in results:
            if record.ok:
                for feature in record.content['results']:
                    content.append(CollectionsFeature(feature))

        return IndexResults(content, results)

    @logging_utils.log_entrance_exit
    def remove_image(self, collection_id, names):
        """
        Remove images from a collection.

        Args:
            collection_id (str): Collection ID to remove images from.
            names (str or sequence of strs): List of image names to be removed.

        Returns:
            :class:`RemoveImageResults <helios.collections_api.RemoveImageResults>`

        """
        if not isinstance(names, (list, tuple)):
            names = [names]

        # Create messages for worker.
        Message = namedtuple('Message', ['collection_id', 'img_name'])
        messages = [Message(collection_id, x) for x in names]

        # Process messages using the worker function.
        results = self._process_messages(self.__remove_image_worker, messages)

        return RemoveImageResults([x.content for x in results], results)

    def __remove_image_worker(self, msg):
        """msg must contain collection_id and img_name"""
        query_str = '{}/{}/{}/images/{}'.format(self._base_api_url,
                                                self._core_api,
                                                msg.collection_id,
                                                msg.img_name)

        try:
            resp = self._request_manager.delete(query_str)
        except requests.exceptions.RequestException as e:
            return Record(message=msg, query=query_str, error=e)

        return Record(message=msg, query=query_str, content=resp.json())

    @logging_utils.log_entrance_exit
    def show(self, collection_id, limit=200, marker=None):
        """
        Get the attributes and image list for collections.

        The results will also contain image names available in the collection.
        These are limited to a maximum of 200 per query.

        Args:
            collection_id (str): Collection ID.
            limit (int, optional): Number of image names to be returned with
                each response. Defaults to 200. Max value of 200 is allowed.
            marker (str, optional): Pagination marker. If the marker is an
                exact match to an existing image, the next image after the
                marker will be the first image returned. Therefore, for normal
                linked list pagination, specify the last image name from the
                current response as the marker value in the next request.
                Partial file names may be specified, in which case the first
                matching result will be the first image returned.

        Returns:
            :class:`ShowResults <helios.collections_api.ShowResults>`

        """
        if not isinstance(collection_id, str):
            raise TypeError('Expected collection_id to be a str but found {} '
                            'instead'.format(type(collection_id)))

        params_str = self._parse_query_inputs(dict(limit=limit, marker=marker))
        query_str = '{}/{}/{}?{}'.format(self._base_api_url,
                                         self._core_api,
                                         collection_id,
                                         params_str)

        resp = self._request_manager.get(query_str)

        return ShowResults(resp.json())

    def show_image(self, collection_id,
                   image_names,
                   out_dir=None,
                   return_image_data=False):
        """
        Get images from a collection.

        Args:
            collection_id (str): Collection ID to add images into.
            image_names (str or sequence of strs): Image names.
            out_dir (optional, str): Directory to write images to.  Defaults to
                None.
            return_image_data (optional, bool): If True images will be returned
                as numpy.ndarrays.  Defaults to False.

        Returns:
            :class:`ShowImageResults <helios.collections_api.ShowImageResults>`

        """
        results = super(Collections, self).show_image(collection_id,
                                                      image_names,
                                                      out_dir=out_dir,
                                                      return_image_data=return_image_data)

        content = []
        for record in results:
            if record.ok:
                content.append(record.content)

        return ShowImageResults(content, results)

    @logging_utils.log_entrance_exit
    def update(self, collections_id, name=None, description=None, tags=None):
        """
        Update a collection.

        Args:
            collections_id (str): Collection ID.
            name (str, optional): Name to be changed to.
            description (str, optional): Description to be changed to.
            tags (str or sequence of strs, optional): Optional comma-delimited
                keyword tags to be changed to.

        """
        if name is None and description is None and tags is None:
            raise ValueError('Update requires at least one keyword argument '
                             'to be used.')

        # need to strip out the Bearer to work with a PATCH for collections
        patch_token = self._request_manager.auth_token['value'].replace('Bearer ', '')

        # Compose parms block
        parms = {}
        if name is not None:
            parms['name'] = name
        if description is not None:
            parms['description'] = description
        if tags is not None:
            if isinstance(tags, (list, tuple)):
                tags = ','.join(tags)
            parms['tags'] = tags
        parms['access_token'] = patch_token

        header = {'name': 'Content-Type',
                  'value': 'application/x-www-form-urlencoded'}

        patch_url = '{}/{}/{}'.format(self._base_api_url,
                                      self._core_api,
                                      collections_id)

        self._request_manager.patch(patch_url, headers=header, data=parms)


class CollectionsFeature(object):
    """
    Individual Collection JSON result.

    Attributes:
        bucket (str): 'bucket' value for the result.
        created_at (str): 'city' value for the result.
        description (str): 'created_at' value for the result.
        id (str): '_id' value for the result.
        images (sequence of strs): 'images' value for the result.
        json (dict): Raw JSON result.
        name (str): 'name' value for the result.
        tags (sequence of strs): 'tags' value for the result.
        updated_at (str): 'updated_at' value for the result.
        user_id (str): 'user_id' value for the result.

    """

    def __init__(self, feature):
        self.json = feature

        # Use dict.get built-in to guarantee all values will be initialized.
        self.bucket = feature.get('bucket')
        self.created_at = feature.get('created_at')
        self.description = feature.get('description')
        self.id = feature.get('_id')
        self.images = feature.get('images')
        self.name = feature.get('name')
        self.tags = feature.get('tags')
        self.updated_at = feature.get('updated_at')
        self.user_id = feature.get('user_id')


class CollectionsFeatureCollection(RecordCollection):
    """Derived class for Collections feature collections."""

    def __init__(self, content, records):
        super(CollectionsFeatureCollection, self).__init__(content, records)

    @property
    def bucket(self):
        """'bucket' values for every result."""
        return [x.bucket for x in self._content]

    @property
    def created_at(self):
        """'city' values for every result."""
        return [x.created_at for x in self._content]

    @property
    def description(self):
        """'created_at' values for every result."""
        return [x.description for x in self._content]

    @property
    def id(self):
        """'_id' values for every result."""
        return [x.id for x in self._content]

    @property
    def json(self):
        """Raw 'json' for every feature."""
        return [x.json for x in self._content]

    @property
    def name(self):
        """'name' values for every result."""
        return [x.name for x in self._content]

    @property
    def tags(self):
        """'tags' values for every result."""
        return [x.tags for x in self._content]

    @property
    def updated_at(self):
        """'updated_at' values for every result."""
        return [x.updated_at for x in self._content]

    @property
    def user_id(self):
        """'user_id' values for every result."""
        return [x.user_id for x in self._content]


class AddImageResults(RecordCollection):
    """Add_image results for Collections API."""

    def __init__(self, content, records):
        super(AddImageResults, self).__init__(content, records)


class IndexResults(CollectionsFeatureCollection):
    """
    Index results for the Collections API.

    IndexResults is an iterable for the results JSON response.  This allows
    the user to iterate and select based on result attributes.

    All features within IndexResults are instances of
    :class:`CollectionsFeature <helios.collections_api.CollectionsFeature>`

    """

    def __init__(self, content, records):
        super(IndexResults, self).__init__(content, records)


class RemoveImageResults(RecordCollection):
    """Remove_image results for the Collections API."""

    def __init__(self, content, records):
        super(RemoveImageResults, self).__init__(content, records)


class ShowImageResults(RecordCollection):
    """
    Show_image results for the Collections API.

    ShowImageResults is an iterable for the fetched image content. Each element
    of the iterable will be an ndarray if return_image_data was True.

    """

    def __init__(self, content, records):
        super(ShowImageResults, self).__init__(content, records)

    @property
    def image_data(self):
        """Image data if return_image_data was True."""
        return self._content

    @property
    def output_file(self):
        """Full paths to all written images."""
        return [x.output_file for x in self._raw if x.ok]

    @property
    def name(self):
        """Names of all images."""
        return [x.name for x in self._raw if x.ok]


class ShowResults(CollectionsFeature):
    """
    Show results for the Collections API.

    The features within ShowResults is an instances of
    :class:`CollectionsFeature <helios.collections_api.CollectionsFeature>`

    """

    def __init__(self, feature):
        super(ShowResults, self).__init__(feature)



"""Helper functions for paths and URLs."""
import hashlib
import os
from datetime import datetime

# Python 2 to 3 fix.
try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse


def parse_time(data):
    """
    Parse time from a URL or image name.

    Args:
        data (str): Image URL or name.
    Returns:
        datetime.datetime: The parsed time as a datetime object.

    """
    time_string = str(os.path.splitext(os.path.split(data)[-1])[0].split('_')[-1])
    time_stamp = datetime.strptime(time_string, '%Y%m%d%H%M%S%f')
    return time_stamp


def parse_camera(data):
    """
    Parse camera name from a URL or image name.

    Args:
        data (str): Image URL or name.
    Returns:
        str: Camera name.

    """

    # Extract tail and remove the extension.
    _, tail = os.path.split(os.path.splitext(data)[0])

    # Split by underscore and remove final index (time string).
    name = '_'.join(tail.split('_')[0:-1])

    # Check for md5 hash (first 4 characters)
    try:
        hash_index = name.index('-')
    except ValueError:
        return name

    # Verify that the first 4 characters are actually the hash.
    md5_hash = hashlib.md5(name[hash_index + 1:].encode('utf-8')).hexdigest()
    if name[0:4] == md5_hash[0:4]:
        return name[5:]
    return name


def parse_image_name(url):
    """
    Parse image name from a URL.

    Args:
        url (str): Image URL.
    Returns:
        str: Image name.

    """
    return os.path.split(url)[-1]


def parse_url(url):
    """
    Parse a URL into its components.

    Args:
        url (str): Image URL.
    Returns:
        urllib.parse.ParseResult: Parsed URL.

    """
    return urlparse(url)



"""
Helios Alerts API.

Methods are meant to represent the core functionality in the developer
documentation.  Some may have additional functionality for convenience.

"""
import logging

from helios.core.mixins import SDKCore, IndexMixin, ShowMixin
from helios.core.structure import RecordCollection


class Alerts(ShowMixin, IndexMixin, SDKCore):
    """
    Helios alerts provide real-time severe weather alerts.

    **National Weather Service:**

    - Severe weather alerts for the United States are provided by the
      National Weather Service. These alerts cover events such as Flood
      Warnings, Severe Thunderstorm Warnings, and Special Weather Statements.

    **Helios:**

    - Alerts generated by Helios are based on the sensor measurements from
      the Observations API. These alerts represent regional areas with a high
      detection confidence and currently include: Road Wetness Watch, Poor
      Visibility Watch, and Heavy Precip Watch.

    """

    _core_api = 'alerts'

    def __init__(self, session=None):
        """
        Initialize Alerts instance.

        Args:
            session (helios.Session object, optional): An instance of the
                Session. Defaults to None. If unused a session will be
                created for you.

        """
        super(Alerts, self).__init__(session=session)
        self._logger = logging.getLogger(__name__)

    def index(self, **kwargs):
        """
        Get alerts matching the provided spatial, text, or
        metadata filters.

        The maximum skip value is 4000. If this is reached, truncated results
        will be returned. You will need to refine your query to avoid this.

        .. _alerts_index_documentation: https://helios.earth/developers/api/alerts/#index

        Args:
            **kwargs: Any keyword arguments found in the alerts_index_documentation_.

        Returns:
             :class:`IndexResults <helios.alerts_api.IndexResults>`

        """
        results = super(Alerts, self).index(**kwargs)

        content = []
        for record in results:
            if record.ok:
                for feature in record.content['features']:
                    content.append(AlertsFeature(feature))

        return IndexResults(content, results)

    def show(self, alert_ids):
        """
        Get attributes for alerts.

        Args:
            alert_ids (str or sequence of strs): Helios alert ID(s).

        Returns:
            :class:`ShowResults <helios.alerts_api.ShowResults>`

        """
        results = super(Alerts, self).show(alert_ids)

        content = []
        for record in results:
            if record.ok:
                content.append(AlertsFeature(record.content))

        return ShowResults(content, results)


class AlertsFeature(object):
    """
    Individual Alert GeoJSON feature.

    Attributes:
        area_description (str): 'areaDesc' value for the feature.
        bbox (sequence of floats): 'bbox' value for the feature.
        category (str): 'category' value for the feature.
        certainty (str): 'certainty' value for the feature.
        country (str): 'country' value for the feature.
        description (str): 'description' value for the feature.
        effective (str): 'effective' value for the feature.
        event (str): 'event' value for the feature.
        expires (str): 'expires' value for the feature.
        headline (str): 'headline' value for the feature.
        id (str): 'id' value for the feature.
        json (dict): Raw 'json' for the feature.
        origin (str): 'origin' value for the feature.
        severity (str): 'severity' value for the feature.
        states (sequence of strs): 'states' value for the feature.
        status (str): 'status' value for the feature.
        urgency (str): 'urgency' value for the feature.

    """

    def __init__(self, feature):
        self.json = feature

        # Use dict.get built-in to guarantee all values will be initialized.
        self.area_description = feature['properties'].get('areaDesc')
        self.bbox = feature.get('bbox')
        self.category = feature['properties'].get('category')
        self.certainty = feature['properties'].get('certainty')
        self.country = feature['properties'].get('country')
        self.description = feature['properties'].get('description')
        self.effective = feature['properties'].get('effective')
        self.event = feature['properties'].get('event')
        self.expires = feature['properties'].get('expires')
        self.headline = feature['properties'].get('headline')
        self.id = feature.get('id')
        self.origin = feature['properties'].get('origin')
        self.severity = feature['properties'].get('severity')
        self.states = feature['properties'].get('states')
        self.status = feature['properties'].get('status')
        self.urgency = feature['properties'].get('urgency')


class AlertsFeatureCollection(RecordCollection):
    """Derived class for Alerts feature collections."""

    def __init__(self, content, records):
        super(AlertsFeatureCollection, self).__init__(content, records)

    @property
    def area_description(self):
        """'areaDesc' values for every feature."""
        return [x.area_description for x in self._content]

    @property
    def bbox(self):
        """'bbox' values for every feature."""
        return [x.bbox for x in self._content]

    @property
    def category(self):
        """'category' values for every feature."""
        return [x['properties']['category'] for x in self._content]

    @property
    def certainty(self):
        """'certainty' values for every feature."""
        return [x.certainty for x in self._content]

    @property
    def country(self):
        """'country' values for every feature."""
        return [x.country for x in self._content]

    @property
    def description(self):
        """'description' values for every feature."""
        return [x.description for x in self._content]

    @property
    def effective(self):
        """'effective' values for every feature."""
        return [x.effective for x in self._content]

    @property
    def event(self):
        """'event' values for every feature."""
        return [x.event for x in self._content]

    @property
    def expires(self):
        """'expires' values for every feature."""
        return [x.expires for x in self._content]

    @property
    def headline(self):
        """'headline' values for every feature."""
        return [x.headline for x in self._content]

    @property
    def id(self):
        """'id' values for every feature."""
        return [x.id for x in self._content]

    @property
    def json(self):
        """Raw 'json' for every feature."""
        return [x.json for x in self._content]

    @property
    def origin(self):
        """'origin' values for every feature."""
        return [x.origin for x in self._content]

    @property
    def severity(self):
        """'severity' values for every feature."""
        return [x.severity for x in self._content]

    @property
    def states(self):
        """'states' values for every feature."""
        return [x.states for x in self._content]

    @property
    def status(self):
        """'status' values for every feature."""
        return [x.status for x in self._content]

    @property
    def urgency(self):
        """'urgency' values for every feature."""
        return [x.urgency for x in self._content]


class IndexResults(AlertsFeatureCollection):
    """
    Index results for the Alerts API.

    IndexResults is an iterable for GeoJSON features.  This allows the
    user to iterate and select based on Feature attributes in each element.

    All features within IndexResults are instances of
    :class:`AlertsFeature <helios.alerts_api.AlertsFeature>`

    """

    def __init__(self, content, records):
        super(IndexResults, self).__init__(content, records)


class ShowResults(AlertsFeatureCollection):
    """
    Show results for the Alerts API.

    ShowResults is an iterable for GeoJSON features.  This allows the
    user to iterate and select based on Feature attributes in each element.

    All features within ShowResults are instances of
    :class:`AlertsFeature <helios.alerts_api.AlertsFeature>`

    """

    def __init__(self, content, records):
        super(ShowResults, self).__init__(content, records)



"""
Helios Observations API.

Methods are meant to represent the core functionality in the developer
documentation.  Some may have additional functionality for convenience.

"""
import logging
import os
from collections import namedtuple
from io import BytesIO

import numpy as np
import pandas as pd
import requests
from PIL import Image
from helios.core.mixins import SDKCore, IndexMixin, ShowMixin
from helios.core.structure import ImageRecord, RecordCollection
from helios.utilities import logging_utils, parsing_utils


class Observations(ShowMixin, IndexMixin, SDKCore):
    """
    The Observations API provides ground-truth data generated by the Helios
    analytics.

    """

    _core_api = 'observations'

    def __init__(self, session=None):
        """
        Initialize Observations instance.

        Args:
            session (helios.Session object, optional): An instance of the
                Session. Defaults to None. If unused a session will be
                created for you.

        """
        super(Observations, self).__init__(session=session)
        self._logger = logging.getLogger(__name__)

    def index(self, **kwargs):
        """
        Get observations matching the provided spatial, text, or
        metadata filters.

        The maximum skip value is 4000. If this is reached, truncated results
        will be returned. You will need to refine your query to avoid this.

        Usage example:

        .. code-block:: python

            import helios
            obs = helios.Observations()
            state = 'Maryland'
            bbox = [-169.352,1.137,-1.690,64.008]
            sensors = 'sensors[visibility][min]=0&sensors[visibility][max]=1'
            results = obs.index(state=state,
                                bbox=bbox,
                                sensors=sensors)

        Usage example for transitions:

        .. code-block:: python

            import helios
            obs = helios.Observations()
            # transition from dry/wet to partial/fully-covered snow roads
            sensors = 'sensors[road_weather][data][min]=6&sensors[road_weather][prev][max]=3'
            results = obs.index(sensors=sensors_query)

        .. _observations_index_documentation: https://helios.earth/developers/api/observations/#index

        Args:
            **kwargs: Any keyword arguments found in the
                observations_index_documentation_.

        Returns:
             :class:`IndexResults <helios.observations_api.IndexResults>`

        """
        results = super(Observations, self).index(**kwargs)

        content = []
        for record in results:
            if record.ok:
                for feature in record.content['features']:
                    content.append(ObservationsFeature(feature))

        return IndexResults(content, results)

    @logging_utils.log_entrance_exit
    def preview(self, observation_ids, out_dir=None, return_image_data=False):
        """
        Get preview images from observations.

        Args:
            observation_ids (str or sequence of strs): list of observation IDs.
            out_dir (optional, str): Directory to write images to.  Defaults to
                None.
            return_image_data (optional, bool): If True images will be returned
                as numpy.ndarrays.  Defaults to False.

        Returns:
            :class:`PreviewResults <helios.observations_api.PreviewResults>`

        """
        if not isinstance(observation_ids, (list, tuple)):
            observation_ids = [observation_ids]

        # Create messages for worker.
        Message = namedtuple('Message', ['observation_id', 'out_dir',
                                         'return_image_data'])
        messages = [Message(x, out_dir, return_image_data) for x in observation_ids]

        # Make sure directory exists.
        if out_dir:
            if not os.path.exists(out_dir):
                os.makedirs(out_dir)

        # Process messages using the worker function.
        results = self._process_messages(self.__preview_worker, messages)

        content = []
        for record in results:
            if record.ok:
                content.append(record.content)

        return PreviewResults(content, results)

    def __preview_worker(self, msg):
        """msg must contain observation_id and check_for_duds"""

        query_str = '{}/{}/{}/preview'.format(self._base_api_url,
                                              self._core_api,
                                              msg.observation_id)

        try:
            resp = self._request_manager.get(query_str)
        except requests.exceptions.RequestException as e:
            return ImageRecord(message=msg, query=query_str, error=e)

        # Parse key from url.
        parsed_url = parsing_utils.parse_url(resp.url)
        _, image_name = os.path.split(parsed_url.path)

        # Read image from response.
        img = Image.open(BytesIO(resp.content))

        # Write image to file.
        if msg.out_dir is not None:
            out_file = os.path.join(msg.out_dir, image_name)
            img.save(out_file)
        else:
            out_file = None

        # Read and return image data.
        if msg.return_image_data:
            img_data = np.array(img)
        else:
            img_data = None

        return ImageRecord(message=msg, query=query_str, name=image_name,
                           content=img_data, output_file=out_file)

    def show(self, observation_ids):
        """
        Get attributes for observations.

        Args:
            observation_ids (str or sequence of strs): Helios observation ID(s).

        Returns:
            :class:`ShowResults <helios.observations_api.ShowResults>`

        """
        results = super(Observations, self).show(observation_ids)

        content = []
        for record in results:
            if record.ok:
                content.append(ObservationsFeature(record.content))

        return ShowResults(content, results)


class ObservationsFeature(object):
    """
    Individual Observation GeoJSON feature.

    Attributes:
        city (str): 'city' value for the feature.
        country (str): 'country' value for the feature.
        description (str): 'description' value for the feature.
        id (str): 'id' value for the feature.
        json (dict): Raw JSON feature.
        prev_id (str): 'prev_id' value for the feature.
        region (str): 'region' value for the feature.
        sensors (dict): 'sensors' value for the feature.
        state (str): 'state' value for the feature.
        time (str): 'time' value for the feature.

    """

    def __init__(self, feature):
        self.json = feature

        # Use dict.get built-in to guarantee all values will be initialized.
        self.city = feature['properties'].get('city')
        self.country = feature['properties'].get('country')
        self.description = feature['properties'].get('description')
        self.id = feature.get('id')
        self.prev_id = feature['properties'].get('prev_id')
        self.region = feature['properties'].get('region')
        self.sensors = feature['properties'].get('sensors')
        self.state = feature['properties'].get('state')
        self.time = feature['properties'].get('time')


class ObservationsFeatureCollection(RecordCollection):
    """Derived class for Observations feature collections."""

    def __init__(self, content, records):
        super(ObservationsFeatureCollection, self).__init__(content, records)

    @property
    def city(self):
        """'city' values for every feature."""
        return [x.city for x in self._content]

    @property
    def country(self):
        """'country' values for every feature."""
        return [x.country for x in self._content]

    @property
    def description(self):
        """'description' values for every feature."""
        return [x.description for x in self._content]

    @property
    def id(self):
        """'id' values for every feature."""
        return [x.id for x in self._content]

    @property
    def json(self):
        """Raw 'json' for every feature."""
        return [x.json for x in self._content]

    @property
    def prev_id(self):
        """'prev_id' values for every feature."""
        return [x.prev_id for x in self._content]

    @property
    def region(self):
        """'region' values for every feature."""
        return [x.region for x in self._content]

    @property
    def sensors(self):
        """'sensors' values for every feature."""
        return [x.sensors for x in self._content]

    @property
    def state(self):
        """'state' values for every feature."""
        return [x.state for x in self._content]

    @property
    def time(self):
        """'time' values for every feature."""
        return [x.time for x in self._content]

    def sensors_to_dataframes(self, output_dir=None, prefix=None):
        """
        Combine sensor blocks and other useful feature information for
        observations into Pandas DataFrame objects.

        DataFrames will contain the time, value, previous value,
        observation ID, and previous observation ID from each feature.

        Optionally, DataFrames can be written to CSV files. These will follow
        the format of {prefix}_{sensor_name}.csv.

        Args:
            output_dir (str, optional): Output directory to write files to. If
                None, then no files will be written. Defaults to None.
            prefix (str, optional): Prefix to append to filenames. If None, no
                prefix will be prepended. Defaults to None.

        Returns:
            dict: Pandas DataFrame objects for each sensor.

        """
        data = {}
        for feature in self._content:
            for sensor, sensor_data in feature.sensors.items():
                if sensor not in data:
                    data[sensor] = []
                data[sensor].append((sensor,
                                     feature.time,
                                     sensor_data.get('data', -1),
                                     sensor_data.get('prev', -1),
                                     feature.id,
                                     feature.prev_id))

        # Establish data frames for each sensor.
        header = ['Sensor', 'Time', 'Data', 'Previous', 'ID', 'Previous_ID']
        output_data = {name: pd.DataFrame(value, columns=header).sort_values(by=['Time'])
                       for name, value in data.items()}

        # If output_dir is specified, write to file.
        if output_dir is not None:
            if prefix is None:
                prefix = ''
            else:
                prefix += '_'
            if not os.path.exists(output_dir):
                os.makedirs(output_dir)

            for sensor_name, df in output_data.items():
                output_file = os.path.join(output_dir,
                                           prefix + sensor_name + '.csv')
                df.to_csv(output_file, na_rep=None, index=False)

        return output_data


class IndexResults(ObservationsFeatureCollection):
    """
    Index results for the Observations API.

    IndexResults is an iterable for GeoJSON features.  This allows the
    user to iterate and select based on Feature attributes.

    All features within IndexResults are instances of
    :class:`ObservationsFeature <helios.observations_api.ObservationsFeature>`

    """

    def __init__(self, content, records):
        super(IndexResults, self).__init__(content, records)


class PreviewResults(RecordCollection):
    """
    Preview results from the Observations API.

    PreviewResults is an iterable for the fetched image content. Each element
    of the iterable will be an ndarray if return_image_data was True.

    """

    def __init__(self, content, records):
        super(PreviewResults, self).__init__(content, records)

    @property
    def image_data(self):
        """Image data if return_image_data was True."""
        return self._content

    @property
    def output_file(self):
        """Full paths to all written images."""
        return [x.output_file for x in self._raw if x.ok]

    @property
    def name(self):
        """Names of all images."""
        return [x.name for x in self._raw if x.ok]


class ShowResults(ObservationsFeatureCollection):
    """
    Show results from the Observations API.

    ShowResults is an iterable for GeoJSON features.  This allows the
    user to iterate and select based on Feature attributes.

    All features within ShowResults are instances of
    :class:`ObservationsFeature <helios.observations_api.ObservationsFeature>`

    """

    def __init__(self, content, records):
        super(ShowResults, self).__init__(content, records)



"""Use the Helios APIs in Python"""
from . import core
from . import utilities
from .alerts_api import Alerts
from .cameras_api import Cameras
from .collections_api import Collections
from .core.session import Session
from .observations_api import Observations
# Configure logger.
from .utilities.logging_utils import configure_logger

configure_logger()

# Cleanup
del configure_logger

__version__ = '2.1.0'



"""Base data structures for the SDK."""


class ContentCollection(object):
    """
    Base class for feature/data results, i.e. content.

    This is a general purpose iterable for content data from the SDK.
    Specific functionality will be defined in child classes.

    Content is a general term for return data from the various API calls
    implemented within the SDK.  For example, index queries return
    GeoJSON feature collections.  Therefore, content will be a list of
    all the returned features within the GeoJSON feature collection.

    Attributes:
        content (sequence): Content that will be the elements in the iterator.
        raw_data (sequence): Raw data for debugging.

    """

    def __init__(self, content, raw_data=None):
        self._content = content
        self._raw = raw_data

    def __delitem__(self, index):
        del self._content[index]

    def __getitem__(self, item):
        return self._content[item]

    def __iter__(self):
        self._idx = 0
        return self

    def __len__(self):
        return len(self._content)

    def __next__(self):
        if self._idx >= self.__len__():
            self._idx = 0
            raise StopIteration
        temp = self.__getitem__(self._idx)
        self._idx += 1
        return temp

    next = __next__  # For Python 2


class RecordCollection(ContentCollection):
    """
    Base class for batch jobs dealing with Records.

    This class is a variation of the FeatureCollection class to work with
    Records. The content attribute will be extracted from each record.  Usage
    is the same as a ContentCollection, but the _raw attribute will give access
    to the underlying Records.

    Attributes:
        content (sequence): Content that will be the elements in the iterator.
        records (sequence of Records): Raw record data for debugging purposes.

    """

    def __init__(self, content, records):
        super(RecordCollection, self).__init__(content, raw_data=records)

    @property
    def failed(self):
        """Records for queries that failed."""
        return [x for x in self._raw if not x.ok]

    @property
    def succeeded(self):
        """Records for queries that succeeded."""
        return [x for x in self._raw if x.ok]

    @property
    def _message(self):
        """
        Messages that were passed to the worker.

        Messages include all the input parameters.

        """
        return [x.message for x in self._raw if x.ok]

    @property
    def _query(self):
        """API query strings that occurred."""
        return [x.query for x in self._raw if x.ok]


class Record(object):
    """
    Record class for general use.

    Args:
        message (tuple): Original message. This will be a namedtuple containing
            all the inputs for an individual call within a batch job.
        query (str): API query.
        content: Returned content. To be defined by method.
        error (exception): Exception that occurred, if any.

    """

    def __init__(self, message=None, query=None, content=None, error=None):
        self.message = message
        self.query = query
        self.content = content
        self.error = error

    @property
    def ok(self):
        """
        Check if failure occurred.

        Returns:
            bool: False if error occurred, and True otherwise.

        """
        if self.error:
            return False
        return True


class ImageRecord(Record):
    """
    Record class for images.

    Args:
        message (tuple): Original message. This will be a namedtuple containing
            all the inputs for an individual call within a batch job.
        query (str): API query.
        content (numpy.ndarray): Image as a Numpy ndarray.
        error (exception): Exception that occurred, if any.
        name (str): Name of image.
        output_file (str): Full path to image file that was written.

    """

    def __init__(self, message=None, query=None, content=None, error=None,
                 name=None, output_file=None):
        super(ImageRecord, self).__init__(message=message, query=query,
                                          content=content, error=error)
        self.name = name
        self.output_file = output_file



"""Manager for the authorization token required to access the Helios API."""
import json
import logging
import os

import requests

# Python 2 compatibility.
try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError


class Session(object):
    """Manages API tokens for authentication.

    Authentication credentials can be specified using the env input parameter,
    environment variables, or a .helios_auth file in your home directory.  See
    the official documentation for more authentication information.

    Required keys:
        - HELIOS_KEY_ID: Client ID from API key pair.
        - HELIOS_KEY_SECRET: Client Secret ID from API key pair.
    Optional keys:
        - HELIOS_API_URL: Optional, URL for API endpoint.

    A session can be established and reused for multiple core API instances.

    .. code-block:: python

        import helios
        sess = helios.Session()
        alerts = helios.Alerts(session=sess)
        cameras = helios.Cameras(session=sess)

    If a session is not specified before hand, one will be initialized
    automatically.  This is less efficient because each core API instance
    will try to initialize a session.

    .. code-block:: python

        import helios
        alerts = helios.Alerts()
        cameras = helios.Cameras()

    """

    token_expiration_threshold = 60  # minutes

    _auth_file = os.path.join(os.path.expanduser('~'), '.helios_auth')
    _default_api_url = r'https://api.helios.earth/v1/'

    def __init__(self, env=None):
        """Initialize Helios Session.

        Args:
            env (dict): Dictionary containing HELIOS_KEY_ID,
                HELIOS_KEY_SECRET, and optionally, 'HELIOS_API_URL'. This will
                override any information in .helios_auth and environment
                variables.

        """
        # Initialize logger
        self._logger = logging.getLogger(__name__)

        # The token will be established with a call to the start_session method.
        self.token = None

        # Try to load essential authentication data from environment or file.
        if env:
            self._logger.info('Using custom env for session.')
            data = env
        elif 'HELIOS_KEY_ID' in os.environ and 'HELIOS_KEY_SECRET' in os.environ:
            self._logger.info('Using environment variables for session.')
            data = os.environ.copy()
        else:
            self._logger.info('Using .helios_auth file for session.')
            try:
                with open(self._auth_file, 'r') as auth_file:
                    data = json.load(auth_file)
            except (IOError, FileNotFoundError):
                raise Exception('No credentials could be found. Be sure to '
                                'set environment variables or setup a '
                                '.helios_auth file.')

        # Extract relevant authentication information from data.
        self._key_id = data['HELIOS_KEY_ID']
        self._key_secret = data['HELIOS_KEY_SECRET']
        try:
            self.api_url = data['HELIOS_API_URL']
        except KeyError:
            self.api_url = self._default_api_url

        # Create token filename based on authentication ID.
        self._token_file = os.path.join(os.path.expanduser('~'),
                                        self._key_id + '.helios_token')

        # Finally, start the session.
        self.start_session()

    def _delete_token(self):
        """Delete token file."""
        try:
            os.remove(self._token_file)
        except (WindowsError, FileNotFoundError):
            pass

    def _get_token(self):
        """
        Gets a fresh token.

        The token will be acquired and then written to file for reuse. If the
        request fails over https, http will be used as a fallback.

        """
        token_url = self.api_url + '/oauth/token'
        try:
            data = {'grant_type': 'client_credentials'}
            auth = (self._key_id, self._key_secret)
            resp = requests.post(token_url, data=data, auth=auth,
                                 verify=True)
            resp.raise_for_status()
        except requests.exceptions.HTTPError:
            self._logger.warning('Getting token over https failed. Falling '
                                 'back to http.')
            token_url_http = 'http' + token_url.split('https')[1]
            data = {'grant_type': 'client_credentials'}
            auth = (self._key_id, self._key_secret)
            resp = requests.post(token_url_http, data=data, auth=auth,
                                 verify=True)

        # If the token cannot be acquired, raise exception.
        resp.raise_for_status()

        token_request = resp.json()
        self.token = {'name': 'Authorization',
                      'value': 'Bearer ' + token_request['access_token']}

        self._write_token_file()

    def _read_token_file(self):
        """Read token from file."""
        with open(self._token_file, 'r') as token_file:
            self.token = json.load(token_file)

    def _write_token_file(self):
        """Write token to file."""
        with open(self._token_file, 'w+') as token_file:
            json.dump(self.token, token_file)

    def start_session(self):
        """
        Begin Helios session.

        This will establish and verify a token for the session.  If a token
        file exists the token will be read and verified. If the token file
        doesn't exist or the token has expired then a new token will be
        acquired.

        """
        try:
            self._read_token_file()
            if not self.verify_token():
                self._get_token()
        except (IOError, FileNotFoundError):
            self._logger.warning('Token file was not found. A new token will '
                                 'be acquired.')
            self._get_token()

    def verify_token(self):
        """
        Verifies if the current token is still valid.

        If the token is bad or if the expiration time is less than the
        threshold False will be returned.

        Returns:
            bool: True if current token is valid, False otherwise.

        """
        resp = requests.get(self.api_url + '/session',
                            headers={self.token['name']: self.token['value']},
                            verify=True)
        resp.raise_for_status()

        json_resp = resp.json()

        if not json_resp['name'] or not json_resp['expires_in']:
            return False

        # Check token expiration time.
        expiration = json_resp['expires_in'] / 60.0
        if expiration < self.token_expiration_threshold:
            self._logger.warning('Token is valid, but expires in %s minutes.',
                                 expiration)
            return False

        self._logger.info('Token is valid for %d minutes.', expiration)
        return True
